#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>

// Some basic notes
#include "notes.h"

#define LED PB0

#define DOT (1 << 7)
#define EIGHTH 1
#define QUARTER 2
#define HALF 4
#define FULL 8

typedef struct {
    uint16_t frequency;
    uint8_t length; // Length of the note (eighth (1), quarter (2), half(4), full(8)), 8th bit is if it is dotted
} Note;

typedef struct {
    Note *notes;
    uint16_t beat_length;
    uint16_t song_length;
} Song;

// Volatile is used to prevent compiler optimizations
const volatile Note jingle_bells[] PROGMEM = {
    {D4, QUARTER}, {B4, QUARTER}, {A4, QUARTER}, {G4, QUARTER}, {D4, HALF | DOT},
    {D4, EIGHTH}, {D4, EIGHTH}, {D4, QUARTER}, {B4, QUARTER}, {A4, QUARTER}, {G4, QUARTER},
    {E4, FULL},
    {E4, QUARTER}, {C5, QUARTER}, {B4, QUARTER}, {A4, QUARTER}, {Fs4_Gb4, FULL},
    {D5, QUARTER}, {D5, QUARTER}, {C5, QUARTER}, {A4, QUARTER}, {B4, FULL},
    {D4, QUARTER}, {B4, QUARTER}, {A4, QUARTER}, {G4, QUARTER}, {D4, FULL},
    {D4, QUARTER}, {B4, QUARTER}, {A4, QUARTER}, {G4, QUARTER}, {E4, HALF | DOT}, {E4, QUARTER},
    {E4, QUARTER}, {C5, QUARTER}, {B4, QUARTER}, {A4, QUARTER}, 
    {D5, QUARTER}, {D5, QUARTER}, {D5, QUARTER}, {D5, QUARTER},
    {E5, QUARTER}, {D5, QUARTER}, {C5, QUARTER}, {A4, QUARTER}, {G4, HALF}, {REST, HALF},

    {B4, QUARTER}, {B4, QUARTER}, {B4, HALF}, {B4, QUARTER}, {B4, QUARTER}, {B4, HALF}, 
    {B4, QUARTER}, {D5, QUARTER}, {G4, QUARTER | DOT}, {A4, EIGHTH}, {B4, FULL},
    {C5, QUARTER}, {C5, QUARTER}, {C5, QUARTER | DOT}, {C5, EIGHTH},
    {C5, QUARTER}, {B4, QUARTER}, {B4, QUARTER}, {B4, EIGHTH}, {B4, EIGHTH},
    {B4, QUARTER}, {A4, QUARTER}, {A4, QUARTER}, {B4, QUARTER}, {A4, HALF}, {D5, HALF},
    {B4, QUARTER}, {B4, QUARTER}, {B4, HALF}, {B4, QUARTER}, {B4, QUARTER}, {B4, HALF},
    {B4, QUARTER}, {D5, QUARTER}, {G4, QUARTER | DOT}, {A4, EIGHTH}, {B4, FULL},
    {C5, QUARTER}, {C5, QUARTER}, {C5, QUARTER | DOT}, {C5, EIGHTH},
    {C5, QUARTER}, {B4, QUARTER}, {B4, QUARTER}, {B4, EIGHTH}, {B4, EIGHTH},
    {D5, QUARTER}, {D5, QUARTER}, {C5, QUARTER}, {A4, QUARTER}, {G4, FULL}

};

const volatile Note deck_the_halls[] PROGMEM = {
    {G4, QUARTER | DOT}, {F4, EIGHTH}, {E4, QUARTER}, {D4, QUARTER}, {C4, QUARTER}, {D4, QUARTER},
    {E4, QUARTER}, {C4, QUARTER}, {D4, EIGHTH}, {E4, EIGHTH}, {F4, EIGHTH}, {D4, EIGHTH}, {E4, QUARTER | DOT},
    {D4, EIGHTH}, {C4, QUARTER}, {B3, QUARTER}, {C4, HALF},

    {G4, QUARTER | DOT}, {F4, EIGHTH}, {E4, QUARTER}, {D4, QUARTER}, {C4, QUARTER}, {D4, QUARTER},
    {E4, QUARTER}, {C4, QUARTER}, {D4, EIGHTH}, {E4, EIGHTH}, {F4, EIGHTH}, {D4, EIGHTH}, {E4, QUARTER | DOT},
    {D4, EIGHTH}, {C4, QUARTER}, {B3, QUARTER}, {C4, HALF},

    {D4, QUARTER | DOT}, {E4, EIGHTH}, {F4, QUARTER}, {D4, QUARTER}, {E4, QUARTER | DOT}, {F4, EIGHTH},
    {G4, QUARTER}, {D4, QUARTER}, {E4, EIGHTH}, {Fs4_Gb4, EIGHTH}, {G4, QUARTER}, {A4, EIGHTH}, {B4, EIGHTH},
    {C5, QUARTER}, {B4, QUARTER}, {A4, QUARTER}, {G4, HALF},

    {G4, QUARTER | DOT}, {F4, EIGHTH}, {E4, QUARTER}, {D4, QUARTER}, {C4, QUARTER}, {D4, QUARTER},
    {E4, QUARTER}, {C4, QUARTER}, {A4, EIGHTH}, {A4, EIGHTH}, {A4, EIGHTH}, {A4, EIGHTH},
    {G4, QUARTER | DOT}, {F4, EIGHTH}, {E4, QUARTER}, {D4, QUARTER}, {C4, HALF}

};

const volatile Note christmas_tree[] PROGMEM = {
    {C4, QUARTER}, {F4, EIGHTH | DOT}, {F4, EIGHTH}, {F4, QUARTER}, {G4, QUARTER}, {A4, EIGHTH | DOT},
    {A4, EIGHTH}, {A4, QUARTER | DOT}, {A4, EIGHTH}, {G4, EIGHTH}, {A4, EIGHTH}, {As4_Bb4, QUARTER},
    {E4, QUARTER}, {G4, QUARTER}, {F4, QUARTER}, {C4, QUARTER},
    
    {F4, EIGHTH | DOT}, {F4, EIGHTH}, {F4, QUARTER}, {G4, QUARTER}, {A4, EIGHTH | DOT},
    {A4, EIGHTH}, {A4, QUARTER | DOT}, {A4, EIGHTH}, {G4, EIGHTH}, {A4, EIGHTH}, {As4_Bb4, QUARTER},
    {E4, QUARTER}, {G4, QUARTER}, {F4, QUARTER | DOT}, {C5, EIGHTH},

    {C5, EIGHTH}, {A4, EIGHTH}, {D5, QUARTER | DOT}, {C5, EIGHTH}, {C5, EIGHTH}, {As4_Bb4, EIGHTH},
    {As4_Bb4, QUARTER | DOT}, {As4_Bb4, EIGHTH}, {As4_Bb4, EIGHTH}, {G4, EIGHTH}, {C5, QUARTER | DOT}, {As4_Bb4, EIGHTH},
    {As4_Bb4, EIGHTH}, {A4, EIGHTH}, {A4, QUARTER}, {C4, QUARTER},

    {F4, EIGHTH | DOT}, {F4, EIGHTH}, {F4, QUARTER}, {G4, QUARTER}, {A4, EIGHTH | DOT}, {A4, EIGHTH},
    {A4, QUARTER | DOT}, {A4, EIGHTH}, {G4, EIGHTH}, {A4, EIGHTH}, {As4_Bb4, QUARTER}, {E4, QUARTER},
    {G4, QUARTER}, {F4, QUARTER}

};

const volatile Note we_wish_you_a_merry_christmas[] PROGMEM = {
    {G4, QUARTER}, {C5, QUARTER}, {C5, EIGHTH}, {D5, EIGHTH}, {C5, EIGHTH}, {B4, EIGHTH},
    {A4, QUARTER}, {A4, QUARTER}, {A4, QUARTER}, {D5, QUARTER}, {D5, EIGHTH}, {E5, EIGHTH}, {D5, EIGHTH},
    {C5, EIGHTH}, {B4, QUARTER}, {G4, QUARTER}, {G4, QUARTER}, 

    {E5, QUARTER}, {E5, EIGHTH}, {F5, EIGHTH}, {E5, EIGHTH}, {D5, EIGHTH}, {C5, QUARTER}, {A4, QUARTER},
    {G4, EIGHTH}, {G4, EIGHTH}, {A4, QUARTER}, {D5, QUARTER}, {B4, QUARTER}, {C5, HALF}, {G4, QUARTER},

    {C5, QUARTER}, {C5, QUARTER}, {C5, QUARTER}, {B4, HALF}, {B4, QUARTER}, {C5, QUARTER}, {B4, QUARTER},
    {A4, QUARTER}, {G4, HALF}, {D5, QUARTER}, 

    {E5, QUARTER}, {D5, QUARTER}, {C5, QUARTER}, {G5, QUARTER}, {G4, QUARTER}, {G4, EIGHTH}, {G4, EIGHTH},
    {A4, QUARTER}, {D5, QUARTER}, {B4, QUARTER}, {C5, HALF}, {G4, QUARTER},

    {C5, QUARTER}, {C5, EIGHTH}, {D5, EIGHTH}, {C5, EIGHTH}, {B4, EIGHTH}, {A4, QUARTER}, {A4, QUARTER}, 
    {A4, QUARTER}, {D5, QUARTER}, {D5, EIGHTH}, {E5, EIGHTH}, {D5, EIGHTH}, {C5, EIGHTH}, {B4, QUARTER},
    {G4, QUARTER}, {G4, QUARTER},

    {E5, QUARTER}, {E5, EIGHTH}, {F5, EIGHTH}, {E5, EIGHTH}, {D5, EIGHTH}, {C5, QUARTER}, {A4, QUARTER},
    {G4, EIGHTH}, {G4, EIGHTH}, {A4, QUARTER}, {D5, QUARTER}, {B4, QUARTER}, {C5, HALF}
};

// Set a frequency in Hz between 61 and 3125
void set_frequency(uint16_t frequency){ // This function is intended for F_CPU=1MHz, results vary for other speeds
    // Bound checking 
    if(frequency > 3125){
        frequency = 3125;
    } else if (frequency < 61){
        frequency = 61;
    }

    // F_CPU is defined in avr-gcc with the "-DF_CPU=1000000" flag
    // 64 is our prescaler value
    uint32_t temp = F_CPU / 64 / frequency;

    OCR1C = (uint8_t) temp; // Our TOP, which sets the frequency

    // Maintain 50% duty cycle
    OCR1A = OCR1C / 2;
}

void enable_PWM_output(){

    TCCR1 |= (1 << COM1A1); // Re-enable OC1A control over PB1

}

void disable_PWM_output(){

    TCCR1 &= ~((1 << COM1A1) | (1 << COM1A0)); // Disconnect OC1A from PB1 to turn it off
    PORTB &= ~(1 << PB1);

}

void play_note(Note n, uint16_t beat_length){
    if(n.frequency != 0){
        enable_PWM_output();
        set_frequency(n.frequency);

        // Flash the LED on each note
        PORTB ^= (1 << LED);
    } else { // If frequency is 0, don't play anything
        disable_PWM_output();
    }

    // The double NOT (!!) tells us if n.length has the DOT bit and returns a 1, otherwise it's a 0
    uint16_t ms = beat_length / 2 * (n.length & ~DOT)
    ms += ms / 2 * !!(n.length & DOT);
    for(uint16_t i = 0; i < ms; i++){
        _delay_ms(1);
    }

    disable_PWM_output();

    // Delay for small period to simulate "Pressing" a new note
    for(uint16_t i = 0; i < beat_length / 16; i++){
        _delay_ms(1);
    }
}

ISR(PCINT0_vect){
    // Handle the interrupt here (nothing to do...)
    // Wakes the CPU up when asleep
}

void init_timer1(){
    // PWM1A to enable PWM (TOP is OCR1C), COM1A1 to output to OC1A
    TCCR1 = (1 << PWM1A) | (1 << COM1A1);

    // Default with TOP being MAX
    OCR1C = 0xFF;

    // 50% duty cycle at startup
    OCR1A = 0xFF / 2;

    // Set PWM pin as an output
    DDRB |= (1 << PB1);

    // Start timer with /64 prescaler (min frequency is ~61 Hz)
    TCCR1 |= (0 << CS13) | (1 << CS12) | (1 << CS11) | (1 << CS10);
}

void sleep(){
    disable_PWM_output();

    // Disable the LED
    PORTB &= ~(1 << LED);

    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
    sleep_enable(); // This doesn't start sleep, but it allows up to go to sleep using sleep_cpu()

    sleep_cpu(); // Put the system to sleep

    sleep_disable(); // Wake up here

    enable_PWM_output();
}

void init_button(){
    // Make our button an input
    DDRB &= ~(1 << PB2);

    // Pin change interrupt on PB2
    GIMSK |= (1 << PCIE); // Enable pin change interrupts
    PCMSK |= (1 << PCINT2); // Use PB2 as our interrupt source
}

int main(){
    // Variable frequency on PB1
    init_timer1();

    init_button();

    sei();

    // Enable the flashing LED
    DDRB |= (1 << LED);

    // Our array of songs
    const Song songs[] = {
        {jingle_bells, 300, sizeof(jingle_bells) / sizeof(Note)},
        {deck_the_halls, 325, sizeof(deck_the_halls) / sizeof(Note)},
        {christmas_tree, 450, sizeof(christmas_tree) / sizeof(Note)},
        {we_wish_you_a_merry_christmas, 325, sizeof(we_wish_you_a_merry_christmas) / sizeof(Note)},
    };

    while(1){
        for(int i = 0; i < sizeof(songs) / sizeof(songs[0]); i++){
            sleep();

            uint16_t beat_length = songs[i].beat_length;
            uint16_t song_length = songs[i].song_length;
            for(int j = 0; j < song_length; j++){
                Note n;

                memcpy_P(&n, &songs[i].notes[j], sizeof(Note));

                play_note(n, beat_length);
            }
        }
    }
}
